# -*- coding: utf-8 -*-
"""
A python library to make psychrometric charts and overlay information in them.

"""
import json
import os
from time import time
from typing import Callable, Union, Dict, Optional, Iterable


PATH_STYLES = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'chart_styles')

DEFAULT_CHART_CONFIG_FILE = os.path.join(
    PATH_STYLES, 'default_chart_config.json')
ASHRAE_CHART_CONFIG_FILE = os.path.join(
    PATH_STYLES, 'ashrae_chart_style.json')
INTERIOR_CHART_CONFIG_FILE = os.path.join(
    PATH_STYLES, 'interior_chart_style.json')
MINIMAL_CHART_CONFIG_FILE = os.path.join(
    PATH_STYLES, 'minimal_chart_style.json')

DEFAULT_ZONES_FILE = os.path.join(
    PATH_STYLES, 'default_comfort_zones.json')

STYLES = {
    "ashrae": ASHRAE_CHART_CONFIG_FILE,
    "default": DEFAULT_CHART_CONFIG_FILE,
    "interior": INTERIOR_CHART_CONFIG_FILE,
    "minimal": MINIMAL_CHART_CONFIG_FILE,
}


def timeit(msg_log: str) -> Callable:
    """Wrapper to time a method call."""
    def real_deco(func) -> Callable:
        def wrapper(*args, **kwargs):
            tic = time()
            out = func(*args, **kwargs)
            print(msg_log + ' TOOK: {:.3f} s'.format(time() - tic))
            return out
        return wrapper
    return real_deco


def _update_config(old_conf: Optional[Dict], new_conf: Dict,
                   verbose: bool=False, recurs_idx: int=0) -> Dict:
    """Update a dict recursively."""
    assert(recurs_idx < 3)
    if old_conf is None:
        return new_conf
    for key, value in old_conf.items():
        if key in new_conf:
            if isinstance(value, dict):
                new_value = _update_config(
                    old_conf[key], new_conf[key], verbose, recurs_idx + 1)
            else:
                new_value = new_conf[key]
                if verbose and new_value != value:
                    print('Update {}: from {} to {}'
                          .format(key, value, new_value))
            old_conf[key] = new_value
    if recurs_idx > 0:
        old_conf.update({key: new_conf[key] for key in filter(
            lambda x: x not in old_conf, new_conf)})
    return old_conf


def _load_config(new_config: Optional[Union[Dict, str]]=None,
                 default_config_file: Optional[str]=None,
                 verbose: bool=False) -> Dict:
    """Load plot parameters from a JSON file."""
    if default_config_file is not None:
        with open(default_config_file) as f:
            config = json.load(f)
    else:
        config = None
    if new_config is not None:
        if isinstance(new_config, str) and new_config.endswith('.json'):
            with open(new_config) as f:
                new_config = json.load(f)
        elif isinstance(new_config, str) and new_config in STYLES:
            with open(STYLES[new_config]) as f:
                new_config = json.load(f)
        else:
            assert(isinstance(new_config, dict))
            new_config = new_config
        config = _update_config(config, new_config, verbose=verbose)

    return config


def load_config(styles: Optional[Union[Dict, str]]=None,
                verbose: bool=False) -> Dict:
    """Load the plot params for the psychrometric chart."""
    return _load_config(
        styles, default_config_file=DEFAULT_CHART_CONFIG_FILE,
        verbose=verbose)


def load_zones(zones: Optional[Union[Dict, str]]=DEFAULT_ZONES_FILE,
               verbose: bool=False) -> Dict:
    """Load a zones JSON file to overlay in the psychrometric chart."""
    return _load_config(zones, verbose=verbose)


def iter_solver(initial_value: float, objective_value: float,
                func_eval: Callable, initial_increment: float=4.,
                num_iters_max: int=100, precision: float=0.01) -> float:
    """Simple solver by iteration."""
    error = 100 * precision
    decreasing = True
    increment = initial_increment
    num_iter = 0
    value_calc = initial_value
    while abs(error) > precision and num_iter < num_iters_max:
        iteration_value = func_eval(value_calc)
        error = objective_value - iteration_value
        if error < 0:
            if not decreasing:
                increment /= 2
                decreasing = True
            increment = max(precision / 10, increment)
            value_calc -= increment
        else:
            if decreasing:
                increment /= 2
                decreasing = False
            increment = max(precision / 10, increment)
            value_calc += increment
        num_iter += 1

        if num_iter >= num_iters_max:
            raise AssertionError(
                'NO CONVERGENCE ERROR AFTER {} ITERATIONS! '
                'Last value: {}, ∆: {}. Objective: {}, iter_value: {}'
                .format(num_iter, value_calc, increment,
                        objective_value, iteration_value))
    return value_calc


def mod_color(color: Iterable, modification: float):
    """Modified Color with an alpha value or a darken/lighten percentage."""
    if abs(modification) < .999:  # is alpha level
        color = list(color) + [modification]
    else:
        color = [max(0., min(1., c * (1 + modification / 100)))
                 for c in color]
    return color



# -*- coding: utf-8 -*-
"""
A python library to make psychrometric charts and overlay information in them.

"""
from psychrochart.chart import PsychroChart


def main():
    """CLI entry point to show the default psychrometric chart."""
    PsychroChart().plot()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
A python library to make psychrometric charts and overlay information in them.

"""
import os
from setuptools import setup, find_packages

from psychrochart import __version__ as version


packages = find_packages(exclude=['docs', '*tests*', 'notebooks', 'htmlcov'])

basedir = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(basedir, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='psychrochart',
    version=version,
    description='A Python 3 library to make psychrometric charts and overlay '
                'information on them.',
    long_description='\n' + long_description,
    keywords='psychrometrics, moist, humid air, climate control, matplotlib',
    author='Eugenio Panadero',
    author_email='eugenio.panadero@gmail.com',
    url='https://github.com/azogue/psychrochart',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],
    packages=packages,
    package_data={
        'psychrochart': ['chart_styles/*.json'],
    },
    install_requires=['numpy>=1.13.1', 'matplotlib>=2.0.2'],
    tests_require=['pytest>=3.0.0'],
)



# -*- coding: utf-8 -*-
"""
A python library to make psychrometric charts and overlay information in them.

Import this module first to set matplotlib to use the Agg backend.
"""
import matplotlib

matplotlib.use('Agg')

# noinspection PyPep8
# noinspection PyUnresolvedReferences
from psychrochart.chart import PsychroChart



# -*- coding: utf-8 -*-
"""
A python library to make psychrometric charts and overlay information in them.

"""
from math import log, exp

from psychrochart.util import iter_solver


PRESSURE_STD_ATM_KPA = 101.325
DELTA_TEMP_C_TO_KELVIN = 273.15

# Eq. (1) 2009 ASHRAE Handbook—Fundamentals (SI)
GAS_CONSTANT_R_DA = 0.287042  # = 8314.472/28.966 kJ/(kg_da·K)
# Eq. (2) 2009 ASHRAE Handbook—Fundamentals (SI)
# GAS_CONSTANT_R_W = 461.524  # = 8314.472/18.015268 = J/(kgw·K)


def pressure_by_altitude(altitude_m: float) -> float:
    """Obtain the standard pressure for a certain sea level.

    Pressure by altitude, eq. (3) 2009 ASHRAE Handbook—Fundamentals (SI).
    """
    return PRESSURE_STD_ATM_KPA * (1 - 2.25577e-5 * altitude_m) ** 5.2559


def water_vapor_pressure(
        w_kg_kga: float, p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> float:
    """Obtain the water vapor pressure from the humidity ratio (w_kg_kga).

    kPa, eq (38)
    2009 ASHRAE Handbook—Fundamentals (SI).
    """
    humid_ratio_vap_pres = .621945
    p_vapor_kpa = p_atm_kpa * w_kg_kga / (humid_ratio_vap_pres + w_kg_kga)
    return p_vapor_kpa


def humidity_ratio(
        p_vapor_kpa: float, p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> float:
    """Obtain the humidity ratio from the water vapor pressure.

    kg water vapor / kg dry air, eq (22)
    2009 ASHRAE Handbook—Fundamentals (SI).
    """
    humid_ratio_vap_pres = .621945
    w_kg_kga = humid_ratio_vap_pres * p_vapor_kpa / (p_atm_kpa - p_vapor_kpa)
    return w_kg_kga


# TODO prec revision:
def humidity_ratio_from_temps(
        dry_bulb_temp_c: float, wet_bulb_temp_c: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> float:
    """Obtain the specific humidity from the dry and wet bulb temperatures.

    kg water vapor / kg dry air, eqs (35) and (37)
    2009 ASHRAE Handbook—Fundamentals (SI).
    """
    w_sat_wet_bulb = humidity_ratio(
        saturation_pressure_water_vapor(wet_bulb_temp_c), p_atm_kpa)
    delta_t = dry_bulb_temp_c - wet_bulb_temp_c
    # assert (delta_t >= 0)
    factor_delta = 1.006 * delta_t
    if dry_bulb_temp_c > 0:
        num_1 = (2501 - 2.326 * wet_bulb_temp_c) * w_sat_wet_bulb
        denom_1 = 2501 + 1.86 * dry_bulb_temp_c - 4.186 * wet_bulb_temp_c
        w_kg_kga = (num_1 - factor_delta) / denom_1
    else:
        num_2 = (2830 - .24 * wet_bulb_temp_c) * w_sat_wet_bulb
        denom_2 = 2830 + 1.86 * dry_bulb_temp_c - 2.1 * wet_bulb_temp_c
        w_kg_kga = (num_2 - factor_delta) / denom_2
    return w_kg_kga


def relative_humidity_from_temps(
        dry_bulb_temp_c: float, wet_bulb_temp_c: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> float:
    """Ratio of the mole fraction of water vapor x_w in a given moist
    air sample to the mole fraction xws in an air sample
    saturated at the same temperature and pressure.

    Eq (24) 2009 ASHRAE Handbook—Fundamentals (SI).
    """
    w_kg_kga = humidity_ratio_from_temps(
        dry_bulb_temp_c, wet_bulb_temp_c, p_atm_kpa)
    p_w_vapor = water_vapor_pressure(w_kg_kga, p_atm_kpa)
    p_sat = saturation_pressure_water_vapor(dry_bulb_temp_c)
    return min(1., p_w_vapor / p_sat)


def specific_volume(
        dry_temp_c: float, p_vapor_kpa: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> float:
    """Obtain the specific volume v of a moist air mixture.

    m3 / kg dry air, eq. (28) 2009 ASHRAE Handbook—Fundamentals (SI).
    """
    w_kg_kga = humidity_ratio(p_vapor_kpa, p_atm_kpa)
    specific_vol_m3_kga = (GAS_CONSTANT_R_DA
                           * (dry_temp_c + DELTA_TEMP_C_TO_KELVIN)
                           * (1 + 1.607858 * w_kg_kga) / p_atm_kpa)
    return specific_vol_m3_kga


def dry_temperature_for_specific_volume_of_moist_air(
        w_kg_kga: float, specific_vol: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> float:
    """Solve the dry bulb temp from humidity ratio and specific volume.

    ºC. Derived from eq. (28), 2009 ASHRAE Handbook—Fundamentals (SI)."""
    return (1741905.36576529 * p_atm_kpa * specific_vol
            / (803929. * w_kg_kga + 500000.) - DELTA_TEMP_C_TO_KELVIN)


def density_water_vapor(p_vapor_kpa: float, dry_bulb_temp_c: float) -> float:
    """Density of water vapor."""
    return (.0022 * p_vapor_kpa * 1000.
            / (dry_bulb_temp_c + DELTA_TEMP_C_TO_KELVIN))


def saturation_pressure_water_vapor(dry_temp_c: float, mode=3) -> float:
    """Saturation pressure of water vapor (kPa) from dry temperature.

    3 approximations:
      - mode 1: ASHRAE formulation
      - mode 2: Simpler, values for T > 0 / T < 0, but same speed as 1
      - mode 3: More simpler, near 2x vs mode 1.
    """
    if mode == 1:  # 2009 ASHRAE Handbook—Fundamentals (SI)
        abs_temp = dry_temp_c + DELTA_TEMP_C_TO_KELVIN
        if dry_temp_c > 0:  # Eq (6) 2009 ASHRAE Handbook—Fundamentals (SI)
            c1 = -5.8002206e3
            c2 = 1.3914993
            c3 = -4.8640239e-2
            c4 = 4.1764768e-5
            c5 = -1.4452093e-8
            c6 = 6.5459673
            ln_p_ws_pa = (c1 / abs_temp + c2 + c3 * abs_temp
                          + c4 * abs_temp ** 2 + c5 * abs_temp ** 3
                          + c6 * log(abs_temp))
        else:  # Eq (5) 2009 ASHRAE Handbook—Fundamentals (SI)
            c7 = -5.6745359e3
            c8 = 6.3925247
            c9 = -9.6778430e-3
            c10 = 6.2215701e-7
            c11 = 2.0747825e-9
            c12 = -9.4840240e-13
            c13 = 4.1635019
            ln_p_ws_pa = (c7 / abs_temp + c8 + c9 * abs_temp
                          + c10 * abs_temp ** 2 + c11 * abs_temp ** 3
                          + c12 * abs_temp ** 4 + c13 * log(abs_temp))
        return exp(ln_p_ws_pa) / 1000
    elif mode == 2:
        if dry_temp_c > 0:
            return 610.5 * exp(
                17.269 * dry_temp_c / (237.3 + dry_temp_c)) / 1000.
        else:
            return 610.5 * exp(
                21.875 * dry_temp_c / (265.5 + dry_temp_c)) / 1000.
    else:  # mode = 3
        return 19314560 * 10. ** (-1779.75 / (237.3 + dry_temp_c))


def enthalpy_moist_air(
        dry_temp_c: float, p_vapor_kpa: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> float:
    """Moist air specific enthalpy.

    KJ / kg. Eqs. (32), (30), (31) 2009 ASHRAE Handbook—Fundamentals (SI)."""
    w_kg_kga = humidity_ratio(p_vapor_kpa, p_atm_kpa)

    c_pa = 1.006  # kJ/(kg·ºC), sensible
    c_pv = 1.86  # kJ/(kg·ºC), latent
    # Evaporation heat (using constant value. For water at 0ºC, it's 2501)
    # constant_water_evaporation_heat = 2503  # KJ / kg
    constant_water_evaporation_heat = 2501  # KJ / kg

    h_a = c_pa * dry_temp_c
    h_v = w_kg_kga * (constant_water_evaporation_heat + c_pv * dry_temp_c)
    h_m = h_a + h_v
    return h_m


def dry_temperature_for_enthalpy_of_moist_air(
        w_kg_kga: float, enthalpy: float) -> float:
    """Solve the dry bulb temp from humidity ratio and specific enthalpy.

    ºC. Derived from eq. (32), 2009 ASHRAE Handbook—Fundamentals (SI)."""
    return 500. * (enthalpy - 2501.0 * w_kg_kga) / (930.0 * w_kg_kga + 503.0)


def dew_point_temperature(p_w_kpa: float) -> float:
    """Dew point temperature.

    Eqs. (39) and (40) Peppers 1988
    2009 ASHRAE Handbook—Fundamentals (SI).
    """
    try:
        alpha = log(p_w_kpa)
    except ValueError:  # pragma: no cover
        raise AssertionError("Bad water vapor pressure! ({} kPa)"
                             .format(p_w_kpa))
    c14 = 6.54
    c15 = 14.526
    c16 = .7389
    c17 = .009486
    c18 = .4569
    dew_point = (c14 + c15 * alpha + c16 * alpha ** 2 + c17 * alpha ** 3
                 + c18 * p_w_kpa ** .1984)
    if dew_point < 0:
        # print('BAD dew_point: ', dew_point)
        dew_point = 6.09 + 12.608 * alpha + .4959 * alpha ** 2
        # print('GOOD dew_point: ', dew_point)
    return dew_point


def wet_bulb_temperature(
        dry_temp_c: float, relative_humid: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA,
        num_iters_max=100, precision=0.00001) -> float:
    """Wet bulb temperature.

    Requires trial-and-error or numerical solution method
    applying eqs. (23) and (35) or (37)
    2009 ASHRAE Handbook—Fundamentals (SI).
    """
    out = iter_solver(
        dry_temp_c - 10, relative_humid,
        lambda x: relative_humidity_from_temps(
            dry_temp_c, x, p_atm_kpa=p_atm_kpa),
        initial_increment=0.5,
        num_iters_max=num_iters_max, precision=precision)

    # Wet bulb temperature can't be greater than dry bulb temp:
    if out > dry_temp_c:
        out = dry_temp_c
    return out

# def degree_of_saturation(w_kg_kga, wsat_kg_kga):
#     """Ratio of air humidity ratio to humidity ratio of saturated moist air.
#
#     µ (no dimension), eq (12) 2009 ASHRAE Handbook—Fundamentals (SI).
#     """
#     return w_kg_kga / wsat_kg_kga
#
#
# def relative_humidity(p_vapor_kpa, p_sat_vapor_kpa):
#     """Ratio of the mole fraction of water vapor x_w in a given moist
#     air sample to the mole fraction xws in an air sample
#     saturated at the same temperature and pressure.
#
#     Eq (24) 2009 ASHRAE Handbook—Fundamentals (SI).
#     """
#     return p_vapor_kpa / p_sat_vapor_kpa



# -*- coding: utf-8 -*-
"""
A python library to make psychrometric charts and overlay information in them.

"""
import json
from math import atan2, degrees
from matplotlib import patches
from matplotlib.axes import Axes
from matplotlib.path import Path
import matplotlib.pyplot as plt
import numpy as np
from typing import Iterable, Callable, Union, Dict, Optional, AnyStr

from psychrochart.equations import (
    PRESSURE_STD_ATM_KPA, pressure_by_altitude, humidity_ratio,
    specific_volume, dew_point_temperature, water_vapor_pressure,
    enthalpy_moist_air, saturation_pressure_water_vapor,
    dry_temperature_for_enthalpy_of_moist_air, relative_humidity_from_temps,
    dry_temperature_for_specific_volume_of_moist_air)
from psychrochart.util import load_config, load_zones, iter_solver, mod_color


PSYCHRO_CURVES_KEYS = [
    'constant_dry_temp_data', 'constant_humidity_data',
    'constant_h_data', 'constant_v_data', 'constant_rh_data',
    'constant_wbt_data', 'saturation']


def _between_limits(x_data: np.array, y_data: np.array,
                    xmin: float, xmax: float,
                    ymin: float, ymax: float) -> bool:
    data_xmin = min(x_data)
    data_xmax = max(x_data)
    data_ymin = min(y_data)
    data_ymax = max(y_data)
    if ((data_ymax < ymin) or (data_xmax < xmin) or
            (data_ymin > ymax) or (data_xmin > xmax)):
        return False
    return True


class PsychroCurve:
    """Object to store a psychrometric curve for plotting."""

    def __init__(self,
                 x_data: Optional[np.array]=None,
                 y_data: Optional[np.array]=None,
                 style: Optional[Dict]=None,
                 type_curve: Optional[AnyStr]=None,
                 limits: Optional[Dict]=None,
                 label: Optional[AnyStr]=None, label_loc: float=.75):
        self.x_data = x_data
        self.y_data = y_data
        self.style = style or {}
        self._type_curve = type_curve
        self._label = label
        self._label_loc = label_loc
        self._limits = limits
        self._is_patch = style is not None and 'facecolor' in style

    def __dict__(self) -> Dict:
        """Return the curve as a dict."""
        return {
            "x_data": self.x_data,
            "y_data": self.y_data,
            "style": self.style,
            "label": self._label}

    def __bool__(self) -> bool:
        """Return the valid existence of the curve."""
        if self.x_data is not None and len(self.x_data) > 1 \
                and self.y_data is not None and len(self.y_data) > 1:
            return True
        return False

    def __repr__(self) -> str:
        """Object string representation."""
        name = 'PsychroZone' if self._is_patch else 'PsychroCurve'
        if self:
            return '<{} {} values (label: {})>'.format(
                name, len(self.x_data), self._label)
        else:
            return '<Empty {} (label: {})>'.format(name, self._label)

    def to_json(self) -> str:
        """Return the curve as a JSON string."""
        data = self.__dict__()
        data['x_data'] = data['x_data'].tolist()
        data['y_data'] = data['y_data'].tolist()
        return json.dumps(data)

    def from_json(self, json_str: AnyStr):
        """Load a curve from a JSON string."""
        data = json.loads(json_str)
        self.x_data = np.array(data['x_data'])
        self.y_data = np.array(data['y_data'])
        self.style = data.get('style')
        self._label = data.get('label')
        return self

    @staticmethod
    def _annotate_label(ax: Axes, label: AnyStr,
                        text_x: float, text_y: float, rotation: float,
                        text_style: Dict):
        if abs(rotation) > 0:
            text_loc = np.array((text_x, text_y))
            text_style['rotation'] = ax.transData.transform_angles(
                np.array((rotation,)), text_loc.reshape((1, 2)))[0]
            text_style['rotation_mode'] = 'anchor'
        ax.annotate(label, (text_x, text_y), **text_style)

    def plot(self, ax: Axes=None) -> Axes:
        """Plot the curve."""
        if ax is None:
            ax = plt.gca()

        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        if not _between_limits(self.x_data, self.y_data,
                               xmin, xmax, ymin, ymax):
            print('{} (label:{}) Not between limits ([{}, {}, {}, {}]) '
                  '-> x:{}, y:{}'
                  .format(self._type_curve, self._label,
                          xmin, xmax, ymin, ymax, self.x_data, self.y_data))
            return ax

        if self._is_patch:
            assert len(self.y_data) > 2
            verts = list(zip(self.x_data, self.y_data))
            codes = ([Path.MOVETO] + [Path.LINETO] * (len(self.y_data) - 2)
                     + [Path.CLOSEPOLY])
            path = Path(verts, codes)
            patch = patches.PathPatch(path, **self.style)
            ax.add_patch(patch)

            if self._label is not None:
                bbox_p = path.get_extents()
                text_x = .5 * (bbox_p.x0 + bbox_p.x1)
                text_y = .5 * (bbox_p.y0 + bbox_p.y1)
                style = {'ha': 'center', 'va': 'center',
                         "backgroundcolor": [1, 1, 1, .4]}
                if 'edgecolor' in self.style:
                    style['color'] = mod_color(self.style['edgecolor'], -25)
                self._annotate_label(ax, self._label,
                                     text_x, text_y, 0, style)
        else:
            ax.plot(self.x_data, self.y_data, **self.style)
            if self._label is not None:
                self.add_label(ax)

        return ax

    def add_label(self, ax: Axes=None,
                  text_label: Optional[AnyStr]=None,
                  va: Optional[str]=None, ha: Optional[str]=None,
                  loc: float=None, **params) -> Axes:
        """Annotate the curve with its label."""
        num_samples = len(self.x_data)
        assert num_samples > 1
        text_style = {'va': 'bottom', 'ha': 'left'}
        loc_f = self._label_loc if loc is None else loc
        label = self._label if text_label is None else text_label

        def _tilt_params(x_data, y_data, idx_0, idx_f):
            delta_x = x_data[idx_f] - self.x_data[idx_0]
            delta_y = y_data[idx_f] - self.y_data[idx_0]
            rotation_deg = degrees(atan2(delta_y, delta_x))
            if delta_x == 0:
                tilt_curve = 1e12
            else:
                tilt_curve = delta_y / delta_x
            return rotation_deg, tilt_curve

        if num_samples == 2:
            xmin, xmax = ax.get_xlim()
            rotation, tilt = _tilt_params(self.x_data, self.y_data, 0, 1)
            if abs(rotation) == 90:
                text_x = self.x_data[0]
                text_y = (self.y_data[0]
                          + loc_f * (self.y_data[1] - self.y_data[0]))
            elif loc_f == 1.:
                if self.x_data[1] > xmax:
                    text_x = xmax
                    text_y = self.y_data[0] + tilt * (xmax - self.x_data[0])
                else:
                    text_x, text_y = self.x_data[1], self.y_data[1]
                label += '    '
                text_style['ha'] = 'right'
            else:
                text_x = self.x_data[0] + loc_f * (xmax - xmin)
                if text_x < xmin:
                    text_x = xmin + loc_f * (xmax - xmin)
                text_y = self.y_data[0] + tilt * (text_x - self.x_data[0])
        else:
            idx = min(num_samples - 2, int(num_samples * loc_f))
            rotation, tilt = _tilt_params(self.x_data, self.y_data,
                                          idx, idx + 1)
            text_x, text_y = self.x_data[idx], self.y_data[idx]
            text_style['ha'] = 'center'

        if 'color' in self.style:
            text_style['color'] = mod_color(self.style['color'], -25)
        if ha is not None:
            text_style['ha'] = ha
        if va is not None:
            text_style['va'] = va
        if params:
            text_style.update(params)

        self._annotate_label(ax, label, text_x, text_y, rotation, text_style)

        return ax


class PsychroCurves:
    """Object to store a list of psychrometric curves for plotting."""

    def __init__(self,
                 curves: Iterable(PsychroCurve),
                 extra: Optional[Dict]=None,
                 family_label: Optional[AnyStr]=None):
        self.curves = curves
        self.family_label = family_label
        self.extra = extra

    def __len__(self) -> int:
        """Return the # of curves."""
        return len(self.curves)

    def __getitem__(self, item) -> PsychroCurve:
        """Get item from the PsychroCurve list."""
        return self.curves[item]

    def __repr__(self) -> str:
        """Object string representation."""
        return '<{} PsychroCurves (label: {})>'.format(
            len(self), self.family_label)

    def plot(self, ax: Axes=None) -> Axes:
        """Plot the family curves."""
        [curve.plot(ax) for curve in self.curves]

        # Curves family labelling
        if self.curves and self.family_label is not None:
            style = self.curves[0].style or {}
            plt.plot([-1], [-1], label=self.family_label,
                     marker='D', markersize=10, **style)

        return ax


def _gen_mat_curves_range_temps(
        func_curve: Callable,
        dbt_min: float, dbt_max: float, increment: float,
        curves_values: list,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> tuple:
    """Generic curve generation in a range of temperatures."""
    temps = np.arange(dbt_min, dbt_max + increment, increment)
    curves = np.zeros((len(temps), len(curves_values)))

    for i, value in enumerate(curves_values):
        curves[:, i] = func_curve(
            temps, value, p_atm_kpa)
    return temps, curves


def _curve_constant_humidity_ratio(
        dry_temps: Iterable, rh_percentage: float=100.,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA, mode_sat=1) -> np.array:
    """Generate a curve (numpy array) of constant humidity ratio."""
    return np.array(
        [1000 * humidity_ratio(
            saturation_pressure_water_vapor(t, mode=mode_sat)
            * rh_percentage / 100., p_atm_kpa)
         for t in dry_temps])


def _make_zone_dbt_rh(
        t_min: float, t_max: float, increment: float,
        rh_min: float, rh_max: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA,
        style: Optional[Dict]=None,
        label: Optional[AnyStr]=None) -> PsychroCurve:
    """Generate points for zone between constant dry bulb temps and RH."""
    temps = np.arange(t_min, t_max + increment, increment)
    curve_rh_up = _curve_constant_humidity_ratio(temps, rh_max, p_atm_kpa)
    curve_rh_down = _curve_constant_humidity_ratio(temps, rh_min, p_atm_kpa)
    abs_humid = np.array(list(curve_rh_up)
                         + list(curve_rh_down[::-1]) + [curve_rh_up[0]])
    temps_zone = np.array(list(temps) + list(temps[::-1]) + [temps[0]])
    return PsychroCurve(temps_zone, abs_humid, style,
                        type_curve='constant_rh_data', label=label)


def _make_zone(
        zone_conf: Dict, increment: float,
        p_atm_kpa: float=PRESSURE_STD_ATM_KPA) -> PsychroCurve:
    """Generate points for zone between constant dry bulb temps and RH."""
    if zone_conf['zone_type'] == 'dbt-rh':
        return _make_zone_dbt_rh(
            *zone_conf['points_x'], increment,
            *zone_conf['points_y'], p_atm_kpa, zone_conf['style'],
            label=zone_conf.get('label'))
    elif zone_conf['zone_type'] == 'xy-points':
        return PsychroCurve(
            zone_conf['points_x'], zone_conf['points_y'], zone_conf['style'],
            type_curve='custom path', label=zone_conf.get('label'))
    # elif zone_conf['zone_type'] == 'dbt-rh-points':
    # make conversion rh -> w


class PsychroChart:
    """Psychrometric chart object handler"""

    def __init__(self,
                 styles: Union[dict, str]=None,
                 zones_file: Union[dict, str]=None):
        """Initialization of the PsychroChart object."""
        self.d_config = {}
        self.figure_params = {}
        self.dbt_min = self.dbt_max = -100
        self.w_min = self.w_max = -1
        self.temp_step = 1.
        self.altitude_m = -1
        self.chart_params = {}
        self.p_atm_kpa = PRESSURE_STD_ATM_KPA
        self.constant_dry_temp_data = None
        self.constant_humidity_data = None
        self.constant_rh_data = None
        self.constant_h_data = None
        self.constant_v_data = None
        self.constant_wbt_data = None
        self.saturation = None
        self.zones = []

        self._axes = None
        self._legend = None
        self._handlers_annotations = []

        self._make_chart_data(styles, zones_file)

    def __repr__(self) -> str:
        """String representation of the PsychroChart object."""
        return '<PsychroChart [{:g}->{:g} °C, {:g}->{:g} gr/kg_da]>'.format(
            self.dbt_min, self.dbt_max, self.w_min, self.w_max)

    @property
    def axes(self) -> Axes:
        """Return the Axes object plotting the chart if necessary."""
        if self._axes is None:
            self.plot()
        assert isinstance(self._axes, Axes)
        return self._axes

    def _make_chart_data(self,
                         styles: Union[dict, str]=None,
                         zones_file: Union[dict, str]=None):
        """Generate the data to plot the psychrometric chart."""
        # Get styling
        config = load_config(styles)
        self.d_config = config
        self.temp_step = config['limits']['step_temp']

        self.figure_params = config['figure']
        self.dbt_min, self.dbt_max = config['limits']['range_temp_c']
        self.w_min, self.w_max = config['limits']['range_humidity_g_kg']
        self.chart_params = config['chart_params']

        # Base pressure
        if config['limits'].get('pressure_kpa') is not None:
            self.p_atm_kpa = config['limits']['pressure_kpa']
        elif config['limits'].get('altitude_m') is not None:
            self.altitude_m = config['limits']['altitude_m']
            self.p_atm_kpa = pressure_by_altitude(self.altitude_m)

        # Dry bulb constant lines (vertical):
        if self.chart_params["with_constant_dry_temp"]:
            step = self.chart_params["constant_temp_step"]
            style = config['constant_dry_temp']
            temps_vl = np.arange(self.dbt_min, self.dbt_max, step)
            heights = [1000 * humidity_ratio(
                saturation_pressure_water_vapor(t),
                p_atm_kpa=self.p_atm_kpa) for t in temps_vl]

            self.constant_dry_temp_data = PsychroCurves(
                [PsychroCurve(np.array([t, t]),
                              np.array([self.w_min, h]), style,
                              type_curve='constant_dry_temp_data',
                              label=None)
                 for t, h in zip(temps_vl, heights)],
                family_label=self.chart_params["constant_temp_label"])

        # Absolute humidity constant lines (horizontal):
        if self.chart_params["with_constant_humidity"]:
            step = self.chart_params["constant_humid_step"]
            style = config['constant_humidity']
            ws_hl = np.arange(self.w_min + step, self.w_max + step / 10, step)
            dew_points = [
                iter_solver(
                    dew_point_temperature(
                        water_vapor_pressure(
                            w / 1000, p_atm_kpa=self.p_atm_kpa)),
                    w / 1000.,
                    lambda x: humidity_ratio(
                        saturation_pressure_water_vapor(x),
                        p_atm_kpa=self.p_atm_kpa),
                    initial_increment=0.25, num_iters_max=100,
                    precision=0.00001)
                for w in ws_hl]

            self.constant_humidity_data = PsychroCurves(
                [PsychroCurve(np.array([t_dp, self.dbt_max]),
                              np.array([w, w]), style,
                              type_curve='constant_humidity_data',
                              label=None)
                 for w, t_dp in zip(ws_hl, dew_points)],
                family_label=self.chart_params["constant_humid_label"])

        # Constant relative humidity curves:
        if self.chart_params["with_constant_rh"]:
            rh_perc_values = self.chart_params["constant_rh_curves"]
            rh_label_values = self.chart_params.get("constant_rh_labels", [])
            label_loc = self.chart_params.get("constant_rh_labels_loc", .85)
            style = config["constant_rh"]
            temps_ct_rh, curves_ct_rh = _gen_mat_curves_range_temps(
                _curve_constant_humidity_ratio,
                self.dbt_min, self.dbt_max, self.temp_step,
                rh_perc_values, p_atm_kpa=self.p_atm_kpa)

            self.constant_rh_data = PsychroCurves(
                [PsychroCurve(
                    temps_ct_rh, curves_ct_rh[:, i], style,
                    type_curve='constant_rh_data',
                    label_loc=label_loc, label='RH {:g} %'.format(rh)
                    if round(rh, 1) in rh_label_values else None)
                    for i, rh in enumerate(rh_perc_values)],
                family_label=self.chart_params["constant_rh_label"])

        # Constant enthalpy lines:
        if self.chart_params["with_constant_h"]:
            step = self.chart_params["constant_h_step"]
            enthalpy_values = np.arange(*self.chart_params["range_h"], step)
            h_label_values = self.chart_params.get("constant_h_labels", [])
            label_loc = self.chart_params.get("constant_h_labels_loc", 1.)
            style = config["constant_h"]
            temps_max_constant_h = [
                dry_temperature_for_enthalpy_of_moist_air(
                    self.w_min / 1000, h)
                for h in enthalpy_values]
            sat_points = [
                iter_solver(
                    dry_temperature_for_enthalpy_of_moist_air(
                        self.w_min / 1000 + 0.1, h),
                    h,
                    lambda x: enthalpy_moist_air(
                        x, saturation_pressure_water_vapor(x),
                        p_atm_kpa=self.p_atm_kpa),
                    initial_increment=15, num_iters_max=100,
                    precision=0.0005)
                for h in enthalpy_values]

            self.constant_h_data = PsychroCurves(
                [PsychroCurve(
                    [t_sat, t_max], [1000 * humidity_ratio(
                        saturation_pressure_water_vapor(t_sat),
                        self.p_atm_kpa), self.w_min], style,
                    type_curve='constant_h_data',
                    label_loc=label_loc, label='{:g} kJ/kg_da'.format(h)
                    if round(h, 3) in h_label_values else None)
                    for t_sat, t_max, h in zip(
                    sat_points, temps_max_constant_h, enthalpy_values)],
                family_label=self.chart_params["constant_h_label"])

        # Constant specific volume lines:
        if self.chart_params["with_constant_v"]:
            step = self.chart_params["constant_v_step"]
            vol_values = np.arange(
                *self.chart_params["range_vol_m3_kg"], step)
            vol_label_values = self.chart_params.get("constant_v_labels", [])
            label_loc = self.chart_params.get("constant_v_labels_loc", 1.)
            style = config["constant_v"]
            temps_max_constant_v = [
                dry_temperature_for_specific_volume_of_moist_air(
                    0, specific_vol, p_atm_kpa=self.p_atm_kpa)
                for specific_vol in vol_values]
            sat_points = [
                iter_solver(
                    t_max - 5,
                    specific_vol,
                    lambda x: specific_volume(
                        x, saturation_pressure_water_vapor(x),
                        p_atm_kpa=self.p_atm_kpa),
                    initial_increment=2, num_iters_max=100,
                    precision=0.00005)
                for t_max, specific_vol in
                zip(temps_max_constant_v, vol_values)]

            self.constant_v_data = PsychroCurves(
                [PsychroCurve(
                    [t_sat, t_max], [1000 * humidity_ratio(
                        saturation_pressure_water_vapor(t_sat),
                        self.p_atm_kpa), 0],
                    style, type_curve='constant_v_data',
                    label_loc=label_loc, label='{:g} m3/kg_da'.format(vol)
                    if round(vol, 3) in vol_label_values else None)
                    for t_sat, t_max, vol in zip(
                    sat_points, temps_max_constant_v, vol_values)],
                family_label=self.chart_params["constant_v_label"])

        # Constant wet bulb temperature lines:
        if self.chart_params["with_constant_wet_temp"]:
            step = self.chart_params["constant_wet_temp_step"]
            wbt_values = np.arange(*self.chart_params["range_wet_temp"], step)
            wbt_label_values = self.chart_params.get(
                "constant_wet_temp_labels", [])
            label_loc = self.chart_params.get(
                "constant_wet_temp_labels_loc", .05)
            style = config["constant_wet_temp"]
            w_max_constant_wbt = [humidity_ratio(
                saturation_pressure_water_vapor(wbt), self.p_atm_kpa)
                for wbt in wbt_values]

            self.constant_wbt_data = PsychroCurves(
                [PsychroCurve(
                    [wbt, self.dbt_max],
                    [1000 * w_max,
                     1000 * humidity_ratio(
                         saturation_pressure_water_vapor(self.dbt_max)
                         * relative_humidity_from_temps(
                             self.dbt_max, wbt, p_atm_kpa=self.p_atm_kpa),
                         p_atm_kpa=self.p_atm_kpa)], style,
                    type_curve='constant_wbt_data',
                    label_loc=label_loc, label='{:g} °C'.format(wbt)
                    if wbt in wbt_label_values else None)
                    for wbt, w_max in zip(wbt_values, w_max_constant_wbt)],
                family_label=self.chart_params["constant_wet_temp_label"])

        # Saturation line:
        if True:
            sat_style = config["saturation"]
            temps_sat_line, w_sat_line = _gen_mat_curves_range_temps(
                _curve_constant_humidity_ratio,
                self.dbt_min, self.dbt_max, self.temp_step,
                [100], p_atm_kpa=self.p_atm_kpa)

            self.saturation = PsychroCurves(
                [PsychroCurve(
                    temps_sat_line, w_sat_line[:, 0], sat_style,
                    type_curve='saturation')])

        # Zones
        if self.chart_params["with_zones"] or zones_file is not None:
            self.append_zones(zones_file)

    def append_zones(self, zones: Union[dict, str]=None):
        """Append zones as patches to the psychrometric chart."""
        if zones is None:
            # load default 'Comfort' zones (Spain RITE)
            d_zones = load_zones()
        else:
            d_zones = load_zones(zones)

        self.zones.append(
            PsychroCurves(
                [_make_zone(zone_conf, self.temp_step, self.p_atm_kpa)
                 for zone_conf in d_zones['zones']]))

    def plot_points_dbt_rh(self, points: Dict, connectors: list=None) -> Dict:
        """Append individual points to the plot."""
        points_plot = {}
        default_style = {'marker': 'o', 'markersize': 10,
                         'color': [1, .8, 0.1, .8], 'linewidth': 0}
        for key, point in points.items():
            plot_params = default_style.copy()
            if isinstance(point, dict):
                plot_params.update(point.get('style', {}))
                plot_params['label'] = point.get('label')
                point = point['xy']
            temp = point[0]
            w_g_ka = 1000. * humidity_ratio(
                point[1] * saturation_pressure_water_vapor(point[0]) / 100,
                p_atm_kpa=self.p_atm_kpa)
            points_plot[key] = [temp], [w_g_ka], plot_params

        if connectors is not None:
            for i, d_con in enumerate(connectors):
                x_start = points_plot[d_con['start']][0][0]
                y_start = points_plot[d_con['start']][1][0]
                x_end = points_plot[d_con['end']][0][0]
                y_end = points_plot[d_con['end']][1][0]
                x_line = [x_start, x_end]
                y_line = [y_start, y_end]
                style = d_con.get('style', points_plot[d_con['start']][2])
                self._handlers_annotations.append(
                    self.axes.plot(
                        x_line, y_line, dash_capstyle='round', **style))
                self._handlers_annotations.append(
                    self.axes.plot(
                        x_line, y_line,
                        color=list(style['color'][:3]) + [.15],
                        lw=50, solid_capstyle='round'))

        for point in points_plot.values():
            self._handlers_annotations.append(
                self.axes.plot(point[0], point[1], **point[2]))

        return points_plot

    def plot_vertical_dry_bulb_temp_line(
            self, temp: float,
            style: Optional[Dict]=None,
            label: Optional[AnyStr]=None,
            reverse: bool=False,
            **label_params):
        # Vertical lines
        w_max = 1000 * humidity_ratio(
            saturation_pressure_water_vapor(temp), self.p_atm_kpa)

        style_curve = style or self.d_config.get("constant_dry_temp")
        path_y = [w_max, self.w_min] if reverse else [self.w_min, w_max]
        curve = PsychroCurve(
            [temp, temp], path_y, style=style_curve)
        curve.plot(self.axes)
        if label is not None:
            curve.add_label(self.axes, label, **label_params)

    def plot_legend(
            self, loc: str='upper left', markerscale: float=.9,
            frameon: bool=True, fancybox: bool=True,
            edgecolor: Union[str, Iterable]='darkgrey', fontsize: float=15.,
            labelspacing: float=1.5, **params):
        """Append a legend to the psychrochart plot."""
        self._legend = self.axes.legend(
            loc=loc, markerscale=markerscale, frameon=frameon,
            edgecolor=edgecolor, fontsize=fontsize, fancybox=fancybox,
            labelspacing=labelspacing, **params)

    def plot(self):
        """Plot the psychrochart and return the matplotlib Axes instance."""
        # Prepare fig & axis
        fig_params = self.figure_params.copy()
        figsize = fig_params.pop('figsize', (16, 9))
        position = fig_params.pop('position', [0.025, 0.075, 0.925, 0.875])
        fontsize = fig_params.pop('fontsize', 10)
        x_style = fig_params.pop('x_axis', {})
        x_style_labels = fig_params.pop('x_axis_labels', {})
        x_style_ticks = fig_params.pop('x_axis_ticks', {})
        y_style = fig_params.pop('y_axis', {})
        y_style_labels = fig_params.pop('y_axis_labels', {})
        y_style_ticks = fig_params.pop('y_axis_ticks', {})
        partial_axis = fig_params.pop('partial_axis', True)

        # Create figure and format axis
        fig = plt.figure(figsize=figsize, dpi=150, frameon=False)
        ax = fig.gca(position=position)
        ax.yaxis.tick_right()
        ax.yaxis.set_label_position("right")
        plt.xlim([self.dbt_min, self.dbt_max])
        plt.ylim([self.w_min, self.w_max])
        plt.grid(False, which='major', axis='both')
        plt.grid(False, which='minor', axis='both')

        # Apply axis styles
        if fig_params['x_label'] is not None:
            style_axis = x_style_labels.copy()
            style_axis['fontsize'] *= 1.2
            plt.xlabel(fig_params['x_label'], **style_axis)
        if fig_params['y_label'] is not None:
            style_axis = y_style_labels.copy()
            style_axis['fontsize'] *= 1.2
            plt.ylabel(fig_params['y_label'], **style_axis)
        if fig_params['title'] is not None:
            plt.title(fig_params['title'],
                      fontsize=fontsize * 1.5, fontweight='bold')

        plt.setp(ax.spines['right'], **y_style)
        plt.setp(ax.spines['bottom'], **x_style)
        if partial_axis:  # Hide left and top axis
            ax.spines['left'].set_visible(False)
            ax.spines['top'].set_visible(False)
        else:
            plt.setp(ax.spines['left'], **y_style)
            plt.setp(ax.spines['top'], **x_style)

        if x_style_ticks:
            ax.tick_params(axis='x', **x_style_ticks)
        if y_style_ticks:
            ax.tick_params(axis='y', **y_style_ticks)

        if self.chart_params.get("with_constant_dry_temp", True):
            step_label = self.chart_params.get(
                "constant_temp_label_step", None)
            if step_label:  # Explicit xticks
                ticks = np.arange(self.dbt_min,
                                  self.dbt_max + step_label / 10, step_label)
                if not self.chart_params.get(
                        "constant_temp_label_include_limits", True):
                    ticks = [t for t in ticks
                             if t not in [self.dbt_min, self.dbt_max]]
                ax.set_xticks(ticks)
                ax.set_xticklabels(
                    ['{:g}'.format(t) for t in ticks], **x_style_labels)
        else:
            ax.set_xticks([])

        if self.chart_params.get("with_constant_humidity", True):
            step_label = self.chart_params.get(
                "constant_humid_label_step", None)
            if step_label:  # Explicit xticks
                ticks = np.arange(self.w_min, self.w_max + step_label / 10,
                                  step_label)
                if not self.chart_params.get(
                        "constant_humid_label_include_limits", True):
                    ticks = [t for t in ticks
                             if t not in [self.w_min, self.w_max]]
                ax.set_yticks(ticks)
                ax.set_yticklabels(
                    ['{:g}'.format(t) for t in ticks], **y_style_labels)
        else:
            ax.set_yticks([])

        # Plot curves:
        [getattr(self, curve_family).plot(ax=ax)
         for curve_family in PSYCHRO_CURVES_KEYS
         if getattr(self, curve_family) is not None]

        # Plot zones:
        [zone.plot(ax=ax) for zone in self.zones]

        # Set the Axes object
        self._axes = ax
        return ax

    def remove_annotations(self):
        """Remove the annotations made in the chart to reuse it."""
        for line in self._handlers_annotations:
            line[0].remove()
        self._handlers_annotations = []

    def save(self, path_dest, **params):
        """Writes the chart to disk."""
        self.axes.get_figure().savefig(path_dest, **params)
